from typing import Generic, Type, TypeVar

from tramp.results import Value, Nothing


V = TypeVar("V")


class DatabaseStatus(Generic[V]):
    Success: "Type[DatabaseSuccessStatus[V]]"
    Exception: "Type[DatabaseExceptionStatus[V]]"


class DatabaseSuccessStatus(Value, DatabaseStatus):
    def __eq__(self, other):
        if not isinstance(other, DatabaseStatus):
            return NotImplemented

        if not isinstance(other, DatabaseSuccessStatus):
            return False

        return self.value == other.value

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.value!r})"


class DatabaseExceptionStatus(DatabaseStatus, Nothing):
    __match_args__ = ("exception",)

    def __new__(cls, *_):
        return object.__new__(cls)

    def __init__(self, exception: Exception) -> None:
        self.exception = exception

    def __eq__(self, other):
        if not isinstance(other, DatabaseStatus):
            raise NotImplementedError()

        if not isinstance(other, DatabaseExceptionStatus):
            return False

        return True

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.exception!r})"


DatabaseStatus.Success = DatabaseSuccessStatus
DatabaseStatus.Exception = DatabaseExceptionStatus



from collections import ChainMap
from typing import Any, TypeVar, Type, cast, MutableMapping
from tramp.results import Result, Value, Nothing
from itertools import zip_longest

T = TypeVar("T")


class FieldMetadata:
    """Base type for all field metadata types. Field metadata instances carry the information about how a field on a
    model should be handled by Ommi and the database driver. The union operator can be used on field metadata instances
    to aggregate them into a single metadata object."""

    metadata: MutableMapping[str, Any]

    def __contains__(self, key: str) -> bool:
        return key in self.metadata

    def __eq__(self, other: "FieldMetadata | Any") -> bool:
        if not isinstance(other, FieldMetadata):
            raise NotImplementedError

        return self.metadata == other.metadata

    def __or__(self, other: "FieldMetadata") -> "FieldMetadata":
        if not isinstance(other, FieldMetadata):
            raise NotImplementedError

        return AggregateMetadata(self, other)

    def __repr__(self):
        return f"{type(self).__name__}({', '.join(f'{k}={v}' for k, v in self.metadata.items())})"

    def get(self, key: str, default: T = None) -> T:
        return self.metadata.get(key, default)

    def matches(self, check_for: "FieldMetadata | Type[FieldMetadata]") -> bool:
        match check_for:
            case type() as metadata_type if issubclass(metadata_type, FieldMetadata):
                return isinstance(self, metadata_type)

            case FieldMetadata() as metadata:
                return self == metadata

            case _:
                return False


class AggregateMetadata(FieldMetadata):
    """Aggregates field metadata instances (usually using the union operator) so a field can have multiple kinds of
    metadata applied simply."""

    metadata: ChainMap[str, Any]

    def __init__(self, *fields: "FieldMetadata") -> None:
        self._fields = []
        self.metadata = ChainMap({})

        for field in fields:
            self._add_field(field)

    def __eq__(self, other: "AggregateMetadata | Any") -> bool:
        if not isinstance(other, AggregateMetadata):
            raise NotImplementedError

        return all(a == b for a, b in zip_longest(self._fields, other._fields))

    def __or__(self, other: "FieldMetadata") -> "FieldMetadata":
        if not isinstance(other, FieldMetadata):
            raise NotImplementedError

        self._add_field(other)
        return self

    def _add_field(self, field: "FieldMetadata") -> None:
        """Adds field metadata to the aggregate metadata. This is done in a non-destructive way so as to prevent changes
        made to the aggregate metadata mapping from propagating to the the aggregated field metadata instances."""
        self._fields.append(field)
        self.metadata.maps.append(field.metadata)

    def matches(self, metadata: "FieldMetadata | Type[FieldMetadata]") -> bool:
        return any(f.matches(metadata) for f in self._fields)


class MetadataFlag(FieldMetadata):
    def __eq__(self, other: "MetadataFlag | Any") -> bool:
        if not isinstance(other, MetadataFlag):
            raise NotImplementedError

        return self.metadata == other.metadata


class FieldType(FieldMetadata):
    """Field metadata type for setting the field's data type."""

    def __init__(self, field_type: Any):
        self.metadata = {"field_type": field_type}


class StoreAs(FieldMetadata):
    """Field metadata type for setting the name to use when passing the field to the database backend."""

    def __init__(self, store_as: str):
        self.metadata = {"store_as": store_as}


def create_metadata_type(
    name: str, metadata_type: "Result[Type[FieldMetadata]]" = Nothing(), /, **kwargs
) -> Type[FieldMetadata]:
    """Helper function for creating a simple field metadata type that has preloaded default values. Useful for creating
    field types that don't need values set at creation."""
    return cast(
        Type[FieldMetadata],
        type(name, (metadata_type.value_or(FieldMetadata),), {"metadata": kwargs}),
    )


def create_metadata_flag(name: str) -> FieldMetadata:
    """Helper function for creating field metadata flag instances. These flags are field metadata instances that contain
    a singular boolean True value to indicate that they're set. Flags should be treated as singletons for purposes of
    identity checks."""
    return create_metadata_type(name, Value(MetadataFlag), **{f"__flag_{name}": True})()


Auto = create_metadata_flag("Auto")
Key = create_metadata_flag("Key")



from types import MethodType
from typing import Callable, ParamSpec, TypeVar

try:
    from typing import Self
except ImportError:
    from typing import Any as Self

P = ParamSpec("P")
R = TypeVar("R")


class ContextualMethod:
    def __init__(self, method: Callable[P, R]):
        self._method = method
        self._classmethod = method

    def classmethod(self, method: Callable[P, R]) -> Self:
        if method.__name__ != self._method.__name__:
            raise ValueError(
                f"Method name {method.__name__!r} does not match {self._method.__name__!r}"
            )

        self._classmethod = method
        return self

    def __get__(self, instance, owner) -> Callable[P, R]:
        if instance is None:
            return MethodType(self._classmethod, owner)

        return MethodType(self._method, instance)


def contextual_method(func: Callable[P, R]) -> ContextualMethod:
    return ContextualMethod(func)



from abc import ABC, abstractmethod
from functools import wraps
from typing import (
    Type,
    TypeAlias,
    TypeVar,
    Callable,
    Awaitable,
    Generator,
    Any,
    Generic,
    ParamSpec,
)

import ommi.driver_context
import ommi.model_collections
from ommi.models import OmmiModel
from ommi.query_ast import ASTGroupNode
from ommi.statuses import DatabaseStatus


DriverName: TypeAlias = str
DriverNiceName: TypeAlias = str
M = TypeVar("M")
T = TypeVar("T")
P = ParamSpec("P")


class DatabaseAction(Generic[T]):
    """A coroutine wrapper that wraps the coroutine's return in a DatabaseStatus result. To force exceptions to be
    raised there is an or_raise coroutine that can be awaited.

        result = await driver.fetch(predicate)  # Can be a DatabaseStatus.Success or a DatabaseStatus.Exception
        ...
        result = await driver.fetch(predicate).or_raise()  # DatabaseStatus.Success or raises the exception
    """

    def __init__(self, awaitable: Awaitable[T]):
        self._awaitable = awaitable

    def __await__(self) -> Generator[Any, None, DatabaseStatus[T]]:
        return self._run().__await__()

    async def or_raise(self) -> DatabaseStatus[T]:
        return DatabaseStatus.Success(await self._awaitable)

    async def _run(self) -> DatabaseStatus[T]:
        try:
            result = await self._awaitable

        except Exception as error:
            return DatabaseStatus.Exception(error)

        else:
            return DatabaseStatus.Success(result)


def database_action(func: Callable[P, T]) -> Callable[P, DatabaseAction[T]]:
    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> DatabaseAction[T]:
        return DatabaseAction[T](func(*args, **kwargs))

    return wrapper


class DriverConfig:
    pass


class AbstractDatabaseDriver(ABC):
    __drivers__: "dict[DriverName, Type[AbstractDatabaseDriver]]"
    driver_name: DriverName

    @classmethod
    @abstractmethod
    def add_driver(cls, driver: "Type[AbstractDatabaseDriver]"):
        ...

    @classmethod
    @abstractmethod
    def disable_driver(cls, name: DriverName):
        ...

    @property
    @abstractmethod
    def connected(self) -> bool:
        ...

    @abstractmethod
    async def add(self, *items: M) -> DatabaseStatus[M]:
        ...

    @abstractmethod
    async def connect(self) -> DatabaseStatus:
        ...

    @abstractmethod
    async def count(self, *predicates: ASTGroupNode | Type[M]) -> DatabaseStatus[int]:
        ...

    @abstractmethod
    async def delete(self, *items: OmmiModel) -> DatabaseStatus:
        ...

    @abstractmethod
    async def disconnect(self) -> DatabaseStatus:
        ...

    @abstractmethod
    async def fetch(
        self, *predicates: ASTGroupNode | Type[OmmiModel]
    ) -> DatabaseStatus:
        ...

    @abstractmethod
    async def sync_schema(
        self, models: "ommi.model_collections.ModelCollection | None"
    ) -> DatabaseStatus:
        ...

    @abstractmethod
    async def update(self, *items: OmmiModel) -> DatabaseStatus:
        ...


class DatabaseDriver(AbstractDatabaseDriver, ABC):
    __drivers__ = {}
    driver_name: DriverName
    nice_name: DriverNiceName
    auto_value_factories: dict[Type[T], Callable[[OmmiModel], T]] = {}

    def __init_subclass__(cls, **kwargs):
        cls.driver_name = kwargs.pop(
            "driver_name", getattr(cls, "driver_name", cls.__name__)
        )
        cls.nice_name = kwargs.pop("nice_name", getattr(cls, "nice_name", cls.__name__))

        super().__init_subclass__(**kwargs)
        cls.add_driver(cls)

    def __init__(self, config: DriverConfig):
        super().__init__()
        self.config = config

    @classmethod
    def add_driver(cls, driver: "Type[AbstractDatabaseDriver]"):
        cls.__drivers__[driver.driver_name] = driver

    @classmethod
    def disable_driver(cls, name: DriverName):
        cls.__drivers__.pop(name, None)

    def __enter__(self):
        if hasattr(self, "_driver_context"):
            return self

        self._driver_context = ommi.driver_context.UseDriver(self)
        return self._driver_context.__enter__()

    def __exit__(self, *args):
        if not hasattr(self, "_driver_context"):
            return

        context = self._driver_context
        del self._driver_context
        try:
            return context.__exit__(*args)
        except ValueError:
            return True

    async def __aenter__(self):
        await self.connect().or_raise()
        self.__enter__()
        return self

    async def __aexit__(self, *args):
        await self.disconnect().or_raise()
        self.__exit__(*args)
        return



from dataclasses import dataclass, field as dc_field
from inspect import get_annotations
from typing import Callable, overload, Type, TypeVar, Any, Generator, get_origin, Annotated, get_args, Awaitable
from tramp.results import Result

import ommi.drivers as drivers
import ommi.query_ast as query_ast
from ommi.field_metadata import FieldMetadata, AggregateMetadata, FieldType, StoreAs, create_metadata_type
from ommi.statuses import DatabaseStatus
from ommi.contextual_method import contextual_method
from ommi.driver_context import active_driver
import ommi.model_collections


T = TypeVar("T", bound=Type)

DRIVER_DUNDER_NAME = "__ommi_driver__"
MODEL_NAME_DUNDER_NAME = "__ommi_model_name__"
MODEL_NAME_CLASS_PARAM = "name"
METADATA_DUNDER_NAME = "__ommi_metadata__"


_global_collection = None


def get_global_collection() -> "ommi.model_collections.ModelCollection":
    global _global_collection
    if not _global_collection:
        _global_collection = ommi.model_collections.ModelCollection()

    return _global_collection


class QueryableFieldDescriptor:
    def __init__(self, name, field):
        self.field = field
        self.name = name

    def __get__(self, instance, owner):
        if instance is None:
            return query_ast.ASTReferenceNode(self, owner)

        return self.field


@dataclass
class OmmiField:
    name: str
    annotation: Any
    default: Any | None = None


@dataclass
class OmmiMetadata:
    model_name: str
    fields: dict[str, OmmiField]
    collection: "ommi.model_collections.ModelCollection" = dc_field(
        default_factory=get_global_collection
    )

    def clone(self, **kwargs) -> "OmmiMetadata":
        return OmmiMetadata(
            **{name: kwargs.get(name, value) for name, value in vars(self).items()}
        )


def _get_value(
    class_params: dict[str, Any],
    param_name: str,
    cls: Type[Any],
    dunder_name: str,
    default: Any,
) -> Any:
    return class_params.pop(param_name, getattr(cls, dunder_name, default))


class OmmiModel:
    __ommi_metadata__: OmmiMetadata

    @contextual_method
    def get_driver(
        self, driver: "drivers.DatabaseDrivers | None" = None
    ) -> "drivers.DatabaseDriver | None":
        return driver or type(self).get_driver()

    @get_driver.classmethod
    def get_driver(cls, driver: "drivers.DatabaseDrivers | None" = None) -> "drivers.DatabaseDriver | None":
        return driver or active_driver.get(None)

    def add(self) -> "drivers.DatabaseAction[DatabaseStatus[OmmiModel]] | Awaitable[DatabaseStatus[OmmiModel]]":
        return self.get_driver().add(self)

    @contextual_method
    def delete(
        self, driver: "drivers.DatabaseDriver | None" = None
    ) -> "drivers.DatabaseAction[DatabaseStatus[drivers.DatabaseDriver]] | Awaitable[DatabaseStatus[drivers.DatabaseDriver]]":
        return self.get_driver(driver).delete(self)

    @delete.classmethod
    def delete(
        cls, *items: "OmmiModel", driver: "drivers.DatabaseDriver | None" = None
    ) -> "drivers.DatabaseAction[DatabaseStatus[drivers.DatabaseDriver]] | Awaitable[DatabaseStatus[drivers.DatabaseDriver]]":
        return cls.get_driver(driver).delete(*items)

    @classmethod
    def count(
        cls,
        *predicates: "ASTGroupNode | DatabaseModel | bool",
        columns: Any | None = None,
        driver: "drivers.DatabaseDriver | None" = None,
    ) -> "drivers.DatabaseAction[DatabaseStatus[int]] | Awaitable[DatabaseStatus[int]]":
        return cls.get_driver(driver).count(
            cls, *predicates, *cls._build_column_predicates(columns)
        )

    @classmethod
    def fetch(
        cls,
        *predicates: "ASTGroupNode | DatabaseModel | bool",
        driver: "drivers.DatabaseDriver | None" = None,
        **columns: Any,
    ) -> "drivers.DatabaseAction[DatabaseStatus[list[OmmiModel]]] | Awaitable[DatabaseStatus[list[OmmiModel]]]":
        return cls.get_driver(driver).fetch(cls, *predicates, *cls._build_column_predicates(columns))

    def sync(
        self, driver: "drivers.DatabaseDriver | None" = None
    ) -> "drivers.DatabaseAction[DatabaseStatus[drivers.DatabaseDriver]] | Awaitable[DatabaseStatus[drivers.DatabaseDriver]]":
        return self.get_driver(driver).update(self)

    @classmethod
    def _build_column_predicates(
        cls, columns: dict[str, Any]
    ) -> "Generator[query_ast.ASTComparisonNode | bool, None, None]":
        if not columns:
            return

        for name, value in columns.items():
            if name not in cls.__fields__:
                raise ValueError(f"Invalid column {name!r} for model {cls.__name__}")

            yield getattr(cls, name) == value


@overload
def ommi_model(
    cls: None = None, *, collection: "ommi.model_collections.ModelCollection | None" = None
) -> Callable[[T], T | Type[OmmiModel]]:
    ...


@overload
def ommi_model(cls: T) -> T | Type[OmmiModel]:
    ...


def ommi_model(
    cls: T | None = None, /, **kwargs
) -> T | Type[OmmiModel] | Callable[[T], T | Type[OmmiModel]]:
    def wrap_model(c: T) -> T | Type[OmmiModel]:
        model = _create_model(c, **kwargs)
        _register_model(model, Result.Value(kwargs["collection"]) if "collection" in kwargs else Result.Nothing)
        return model

    return wrap_model if cls is None else wrap_model(cls)


def _create_model(c: T, **kwargs) -> T | Type[OmmiModel]:
    metadata_factory = (
        c.__ommi_metadata__.clone
        if hasattr(c, METADATA_DUNDER_NAME)
        else OmmiMetadata
    )

    fields = _get_fields(get_annotations(c))
    return type.__new__(
        type(c),
        f"OmmiModel_{c.__name__}",
        (c, OmmiModel),
        {
            name: QueryableFieldDescriptor(fields[name].get("store_as"), getattr(c, name, None))
            for name in get_annotations(c)
            if not name.startswith("_")
        }
        | {
            METADATA_DUNDER_NAME: metadata_factory(
                model_name=_get_value(
                    kwargs,
                    MODEL_NAME_CLASS_PARAM,
                    c,
                    MODEL_NAME_DUNDER_NAME,
                    c.__name__,
                ),
                fields=fields,
            )
        },
    )


def _register_model(model: Type[OmmiModel], collection: "Result[ommi.model_collections.ModelCollection]"):
    get_collection(collection, model).add(model)


def get_collection(
        collection: "Result[ommi.model_collections.ModelCollection]",
        model: Type[OmmiModel] | None = None,
) -> "ommi.model_collections.ModelCollection":
    return collection.value_or(getattr(model, METADATA_DUNDER_NAME).collection if model else get_global_collection())


def _get_fields(fields: dict[str, Any]) -> dict[str, FieldMetadata]:
    ommi_fields = {}
    for name, hint in fields.items():
        ommi_fields[name] = AggregateMetadata()
        if get_origin(hint) == Annotated:
            field_type, *annotations = get_args(hint)
            _annotations = []
            for annotation in annotations:
                match annotation:
                    case FieldMetadata():
                        ommi_fields[name] |= annotation

                    case _:
                        _annotations.append(annotation)

            if _annotations:
                field_type = Annotated.__class_getitem__(field_type, *_annotations)  # Hack to support 3.10

        else:
            field_type = hint

        ommi_fields[name] |= FieldType(field_type)
        if not ommi_fields[name].matches(StoreAs):
            ommi_fields[name] |= StoreAs(name)

        ommi_fields[name] |= create_metadata_type("FieldMetadata", field_name=name, field_type=field_type)()

    return ommi_fields



from contextvars import ContextVar
from contextlib import contextmanager
from typing import Generator, TypeVar, Generic

import ommi.drivers as drivers


T = TypeVar("T", bound="drivers.DatabaseDriver")


active_driver = ContextVar("active_driver")


class UseDriver(Generic[T]):
    def __init__(self, driver: T):
        self.driver = driver
        self._previous_context_token = None

    def __enter__(self) -> T:
        self._previous_context_token = active_driver.set(self.driver)
        return self.driver

    def __exit__(self, *_):
        active_driver.reset(self._previous_context_token)


def use_driver(driver: T) -> UseDriver[T]:
    return UseDriver(driver)



from enum import auto, Enum
from itertools import zip_longest
from typing import Any

try:
    from typing import Self
except ImportError:
    Self = Any

import ommi.models as models


class ResultOrdering(Enum):
    ASCENDING = auto()
    DESCENDING = auto()


class ASTNode:
    pass


class ASTLogicalOperatorNode(ASTNode, Enum):
    AND = auto()
    OR = auto()


class ASTOperatorNode(ASTNode, Enum):
    EQUALS = auto()
    NOT_EQUALS = auto()
    GREATER_THAN = auto()
    GREATER_THAN_OR_EQUAL = auto()
    LESS_THAN = auto()
    LESS_THAN_OR_EQUAL = auto()


class ASTGroupFlagNode(ASTNode, Enum):
    OPEN = auto()
    CLOSE = auto()


class ASTGroupNode(ASTNode):
    def __init__(
        self, items: "list[ASTComparableNode | ASTLogicalOperatorNode] | None" = None
    ):
        self.items: list[ASTComparableNode | ASTLogicalOperatorNode] = (
            [] if items is None else items
        )
        self.frozen = False
        self.max_results = -1
        self.results_page = 0
        self.sorting: list[ASTReferenceNode] = []

    def __iter__(self):
        self.frozen = True
        if len(self.items) > 1:
            yield ASTGroupFlagNode.OPEN

        try:
            yield from iter(self.items)
        finally:
            self.frozen = False

        if len(self.items) > 1:
            yield ASTGroupFlagNode.CLOSE

    def add(self, item, logical_type=ASTLogicalOperatorNode.AND):
        if self.frozen:
            return

        if isinstance(item, type) and issubclass(item, models.OmmiModel):
            item = ASTReferenceNode(None, item)

        elif len(self.items) > 0 and getattr(self.items[~0], "field", True) is not None:
            self.items.append(logical_type)

        self.items.append(item)

    def limit(self, limit: int, page: int = 0) -> Self:
        self.max_results = limit
        self.results_page = page
        return self

    def sort(self, *on_fields: "ASTReferenceNode") -> Self:
        self.sorting.extend(field for field in on_fields if field not in self.sorting)
        return self

    def __eq__(self, other):
        if not isinstance(other, ASTGroupNode):
            return NotImplemented

        return self._compare_items(other)

    def _compare_items(self, other):
        # Use zip_longest to ensure there aren't additional items in either group
        for items in zip_longest(self, other):
            match items:
                case (ASTLiteralNode(a), ASTLiteralNode(b)) if a == b:
                    continue

                case (
                    ASTReferenceNode(af, am),
                    ASTReferenceNode(bf, bm),
                ) if af == bf and am == bm:
                    continue

                case (
                    ASTComparisonNode(al, ar, ao),
                    ASTComparisonNode(bl, br, bo),
                ) if al._eq(bl) and ar._eq(br) and ao == bo:
                    continue

                case (ASTGroupNode() as a, ASTGroupNode() as b) if a == b:
                    continue

                case (
                    ASTLogicalOperatorNode() as a,
                    ASTLogicalOperatorNode() as b,
                ) if a == b:
                    continue

                case (ASTGroupFlagNode() as a, ASTGroupFlagNode() as b) if a == b:
                    continue

                case _:
                    return False

        return True

    def __repr__(self):
        return f"{type(self).__name__}({self.items!r})"

    def And(self, *comparisons: "SearchGroup | ASTComparisonNode | bool"):
        return self._add_node_or_group(when(*comparisons), ASTLogicalOperatorNode.AND)

    def Or(self, *comparisons: "SearchGroup | ASTComparisonNode | bool"):
        return self._add_node_or_group(when(*comparisons), ASTLogicalOperatorNode.OR)

    def _add_node_or_group(
        self,
        comparison: "SearchGroup | ASTComparisonNode",
        logical_type: ASTLogicalOperatorNode,
    ):
        match comparison:
            case ASTComparisonNode() if len(comparison.group.items) > 1:
                self.add(comparison.group, logical_type)

            case ASTGroupNode() if len(comparison.items) == 1:
                self.add(comparison.items[0], logical_type)

            case _:
                self.add(comparison, logical_type)

        return self


class ASTComparableNode(ASTNode):
    def __init__(self, group):
        self.group = group

    @staticmethod
    def _safe_compare(func):
        def wrapper(self, *args):
            if self.group.frozen:
                return False

            return func(self, *args)

        return wrapper

    @_safe_compare
    def __eq__(self, other) -> "ASTComparisonNode":
        self.group.add(
            node := ASTComparisonNode(self, other, ASTOperatorNode.EQUALS, self.group)
        )
        return node

    @_safe_compare
    def __ne__(self, other) -> "ASTComparisonNode":
        self.group.add(
            node := ASTComparisonNode(
                self, other, ASTOperatorNode.NOT_EQUALS, self.group
            )
        )
        return node

    def __gt__(self, other) -> "ASTComparisonNode":
        self.group.add(
            node := ASTComparisonNode(
                self, other, ASTOperatorNode.GREATER_THAN, self.group
            )
        )
        return node

    def __ge__(self, other) -> "ASTComparisonNode":
        self.group.add(
            node := ASTComparisonNode(
                self, other, ASTOperatorNode.GREATER_THAN_OR_EQUAL, self.group
            )
        )
        return node

    def __lt__(self, other) -> "ASTComparisonNode":
        self.group.add(
            node := ASTComparisonNode(
                self, other, ASTOperatorNode.LESS_THAN, self.group
            )
        )
        return node

    def __le__(self, other) -> "ASTComparisonNode":
        self.group.add(
            node := ASTComparisonNode(
                self, other, ASTOperatorNode.LESS_THAN_OR_EQUAL, self.group
            )
        )
        return node


class ASTReferenceNode(ASTComparableNode):
    __match_args__ = ("field", "model")

    def __init__(self, field, model, ordering=ResultOrdering.ASCENDING):
        super().__init__(ASTGroupNode())
        self._field = field
        self._model = model
        self._ordering = ordering

    def _eq(self, other):
        return self.field == other.field and self.model == other.model

    def __iter__(self):
        yield from (self._field, self._model)

    @property
    def field(self):
        return self._field

    @property
    def model(self):
        return self._model

    @property
    def ordering(self):
        return self._ordering

    @property
    def asc(self) -> "ASTReferenceNode":
        return ASTReferenceNode(self._field, self._model, ResultOrdering.ASCENDING)

    @property
    def desc(self) -> "ASTReferenceNode":
        return ASTReferenceNode(self._field, self._model, ResultOrdering.DESCENDING)

    def __hash__(self):
        return hash((self._field, self._model, self._ordering))

    def __repr__(self):
        return (
            f"<{type(self).__qualname__}"
            f" {self.model.__qualname__ if self.model else None}.{getattr(self._field, 'name', self._field)!r}>"
        )


class ASTLiteralNode(ASTComparableNode):
    __match_args__ = ("value",)

    def __init__(self, value):
        super().__init__(ASTGroupNode())
        self._value = value

    def _eq(self, other):
        if not isinstance(other, ASTLiteralNode):
            return NotImplemented

        return self.value == other.value

    def __iter__(self):
        yield self._value

    @property
    def value(self):
        return self._value

    def __repr__(self):
        return f"{type(self).__name__}({self._value!r})"


class ASTComparisonNode(ASTComparableNode):
    __match_args__ = ("left", "right", "operator")

    def __init__(
        self,
        left: ASTLiteralNode | ASTReferenceNode | Any,
        right: ASTLiteralNode | ASTReferenceNode | Any,
        operator: ASTOperatorNode,
        group=None,
    ):
        super().__init__(group or ASTGroupNode())
        self._left = self._make_node(left)
        self._right = self._make_node(right)
        self._operator = operator

    def __iter__(self):
        yield from (self._left, self._right, self._operator)

    @property
    def left(self):
        return self._left

    @property
    def right(self):
        return self._right

    @property
    def operator(self):
        return self._operator

    def __repr__(self):
        return (
            f"{type(self).__name__}("
            f"{self._left!r}, "
            f"{self._right!r}, "
            f"{self._operator})"
        )

    def And(self, *comparisons: "ASTComparisonNode | SearchGroup | bool"):
        return self.group.And(*comparisons)

    def Or(self, *comparisons: "ASTComparisonNode | SearchGroup | bool"):
        return self.group.Or(*comparisons)

    def _make_node(self, value):
        match value:
            case ASTComparableNode():
                return value

            case _:
                return ASTLiteralNode(value)


def when(
    *comparisons: "ASTComparisonNode | Type[models.DatabaseModel] | bool",
) -> ASTGroupNode:
    match comparisons:
        case (ASTComparisonNode() as comparison,):
            return comparison.group

        case (ASTGroupNode() as group,):
            return group

        case _:
            group = ASTGroupNode()
            for c in comparisons:
                match c:
                    case ASTComparisonNode():
                        group.add(c.group)

                    case _:
                        group.add(c)

            return group



from typing import Type

import ommi.models


class ModelCollection:
    def __init__(self):
        self.models = set()

    def add(self, model: "Type[ommi.models.OmmiModel]"):
        self.models.add(model)



import sqlite3
from dataclasses import dataclass, field as dc_field
from datetime import datetime, date
from typing import Type, Any, TypeVar, Callable, get_origin, Generator

from tramp.results import Result

from ommi.drivers import DatabaseDriver, DriverConfig, database_action
from ommi.model_collections import ModelCollection
from ommi.models import OmmiField, OmmiModel, get_collection
from ommi.query_ast import (
    ASTGroupNode,
    when,
    ASTReferenceNode,
    ASTLiteralNode,
    ASTLogicalOperatorNode,
    ASTComparisonNode,
    ASTOperatorNode,
    ASTGroupFlagNode,
    ResultOrdering,
)

T = TypeVar("T")


@dataclass
class SelectQuery:
    limit: int = 0
    model: Type[OmmiModel] | None = None
    offset: int = 0
    order_by: dict[str, ResultOrdering] = dc_field(default_factory=dict)
    tables: list[OmmiModel] = dc_field(default_factory=list)
    values: list[Any] = dc_field(default_factory=list)
    where: str = ""


@dataclass
class SQLiteConfig(DriverConfig):
    filename: str


class SQLiteDriver(DatabaseDriver, driver_name="sqlite", nice_name="SQLite"):
    config: SQLiteConfig

    logical_operator_mapping = {
        ASTLogicalOperatorNode.AND: "AND",
        ASTLogicalOperatorNode.OR: "OR",
    }

    operator_mapping = {
        ASTOperatorNode.EQUALS: "=",
        ASTOperatorNode.NOT_EQUALS: "!=",
        ASTOperatorNode.GREATER_THAN: ">",
        ASTOperatorNode.GREATER_THAN_OR_EQUAL: ">=",
        ASTOperatorNode.LESS_THAN: "<",
        ASTOperatorNode.LESS_THAN_OR_EQUAL: "<=",
    }

    type_validators = {
        datetime: lambda value: value,  # No-op to avoid type conflict with date
        date: lambda value: value.split()[0],
    }

    type_mapping = {
        int: "INTEGER",
        str: "TEXT",
        float: "REAL",
        bool: "INTEGER",
    }

    def __init__(self, *args):
        super().__init__(*args)
        self._connected = False
        self._db: sqlite3.Connection | None = None

    @property
    def connected(self) -> bool:
        return self._connected

    @database_action
    async def connect(self) -> "SQLiteDriver":
        self._db = sqlite3.connect(self.config.filename)
        self._connected = True
        return self

    @database_action
    async def disconnect(self) -> "SQLiteDriver":
        self._db.close()
        self._connected = False
        return self

    @database_action
    async def add(self, *items: OmmiModel) -> "SQLiteDriver":
        session = self._db.cursor()
        try:
            for item in items:
                self._insert(item, session)
                self._sync_with_last_inserted(item, session)

        except:
            self._db.rollback()
            raise

        else:
            return self

        finally:
            session.close()

    @database_action
    async def count(self, *predicates: ASTGroupNode | Type[OmmiModel]) -> int:
        ast = when(*predicates)
        session = self._db.cursor()
        return self._count(ast, session)

    @database_action
    async def delete(self, *items: OmmiModel) -> "SQLiteDriver":
        models = {}
        for item in items:
            models.setdefault(type(item), []).append(item)

        session = self._db.cursor()
        try:
            for model, items in models.items():
                self._delete_rows(model, items, session)

        except:
            self._db.rollback()
            raise

        else:
            return self

        finally:
            session.close()

    @database_action
    async def fetch(
        self, *predicates: ASTGroupNode | Type[OmmiModel]
    ) -> list[OmmiModel]:
        ast = when(*predicates)
        session = self._db.cursor()
        result = self._select(ast, session)
        return result

    @database_action
    async def sync_schema(
        self, collection: ModelCollection | None = None
    ) -> "SQLiteDriver":
        session = self._db.cursor()
        models = get_collection(
            Result.Value(collection) if collection else Result.Nothing
        ).models
        try:
            for model in models:
                self._create_table(model, session)

        except:
            self._db.rollback()
            raise

        else:
            return self

        finally:
            session.close()

    @database_action
    async def update(self, *items: OmmiModel) -> "SQLiteDriver":
        models = {}
        for item in items:
            models.setdefault(type(item), []).append(item)

        session = self._db.cursor()
        try:
            for model, items in models.items():
                self._update_rows(model, items, session)

        except:
            self._db.rollback()
            raise

        else:
            return self

        finally:
            session.close()

    def _build_column(self, field: OmmiField, pk: bool) -> str:
        column = [
            field.get("store_as"),
            self._get_sqlite_type(field.get("field_type")),
        ]
        if pk:
            column.append("PRIMARY KEY")

        return " ".join(column)

    def _process_ast(self, ast: ASTGroupNode) -> SelectQuery:
        query = SelectQuery(
            limit=ast.max_results,
            offset=ast.results_page * ast.max_results,
            order_by=self._process_ordering(ast.sorting),
        )
        where = []
        node_stack = [iter(ast)]
        while node_stack:
            match next(node_stack[~0], None):
                case ASTGroupNode() as group:
                    node_stack.append(iter(group))

                case ASTReferenceNode(None, model):
                    query.tables.append(model)
                    query.model = query.model or model

                case ASTReferenceNode(field, model):
                    name = f"{model.__ommi_metadata__.model_name}.{field.name}"
                    where.append(name)
                    query.tables.append(model)
                    query.model = query.model or model

                case ASTLiteralNode(value):
                    where.append("?")
                    query.values.append(value)

                case ASTLogicalOperatorNode() as op:
                    where.append(self.logical_operator_mapping[op])

                case ASTOperatorNode() as op:
                    where.append(self.operator_mapping[op])

                case ASTComparisonNode(left, right, op):
                    node_stack.append(iter((left, op, right)))

                case ASTGroupFlagNode.OPEN:
                    if len(node_stack) > 1:
                        where.append("(")

                case ASTGroupFlagNode.CLOSE:
                    if len(node_stack) > 1:
                        where.append(")")

                case None:
                    node_stack.pop()

                case node:
                    raise TypeError(f"Unexpected node type: {node}")

        query.where = " ".join(where)
        return query

    def _process_ordering(
        self, sorting: list[ASTReferenceNode]
    ) -> dict[str, ResultOrdering]:
        return {
            f"{ref.model.__model_name__}.{ref.field.name}": ref.ordering
            for ref in sorting
        }

    def _create_table(self, model: Type[OmmiModel], session: sqlite3.Cursor):
        pk = self._find_primary_key(model)
        columns = ", ".join(
            self._build_column(field, field.get("store_as") == pk)
            for field in model.__ommi_metadata__.fields.values()
        )
        session.execute(
            f"CREATE TABLE IF NOT EXISTS {model.__ommi_metadata__.model_name} ({columns});"
        )

    def _find_primary_key(self, model: Type[OmmiModel]) -> str:
        if not model.__ommi_metadata__.fields:
            raise Exception(f"No fields defined on {model}")

        fields = list(model.__ommi_metadata__.fields.values())
        if name := next(
            (
                f.get("store_as")
                for f in fields
                if f.get("store_as", f.get("field_name")).lower() == "id"
            ),
            None,
        ):
            return name

        for field in fields:
            if (
                field.get("field_type") is int
                or field.get("store_as").casefold() == "id"
            ):
                return field.get("store_as")

        return next(iter(fields)).get("store_as")

    def _insert(self, item: OmmiModel, session: sqlite3.Cursor):
        fields = list(item.__ommi_metadata__.fields.values())
        data = {
            field.get("store_as"): getattr(item, field.get("field_name"))
            for field in fields
        }
        qs = ", ".join(["?"] * len(data))
        columns = ", ".join(data.keys())
        values = tuple(data.values())
        session.execute(
            f"INSERT INTO {item.__ommi_metadata__.model_name} ({columns}) VALUES ({qs});",
            values,
        )

    def _update_rows(
        self,
        model: Type[OmmiModel],
        items: list[OmmiModel],
        session: sqlite3.Cursor,
    ):
        pk = self._find_primary_key(model)
        fields = list(model.__ommi_metadata__.fields.values())
        for item in items:
            values = (
                getattr(item, field.get("field_name"))
                for field in fields
                if field.get("store_as") != pk
            )
            assignments = ", ".join(
                f"{field.get('store_as')} = ?"
                for field in fields
                if field.get("store_as") != pk
            )
            session.execute(
                f"UPDATE {model.__ommi_metadata__.model_name} SET {assignments} WHERE {pk} = ?;",
                (*values, getattr(item, pk)),
            )

    def _delete_rows(
        self,
        model: Type[OmmiModel],
        items: list[OmmiModel],
        session: sqlite3.Cursor,
    ):
        pk = self._find_primary_key(model)
        keys = [getattr(item, pk) for item in items]
        qs = ", ".join(["?"] * len(items))
        session.execute(
            f"DELETE FROM {model.__ommi_metadata__.model_name} WHERE {pk} IN ({qs});",
            keys,
        )

    def _select(self, predicates: ASTGroupNode, session: sqlite3.Cursor):
        query = self._process_ast(predicates)
        query_str = self._build_select_query(query)
        session.execute(query_str, query.values)
        result = session.fetchall()
        return [
            query.model(*self._validate_row_values(query.model, row)) for row in result
        ]

    def _count(self, predicates: ASTGroupNode, session: sqlite3.Cursor):
        query = self._process_ast(predicates)
        query_str = self._build_count_query(query)
        session.execute(query_str, query.values)
        result = session.fetchone()
        return result[0]

    def _validate_row_values(
        self, model: Type[OmmiModel], row: tuple[Any]
    ) -> Generator[Any, None, None]:
        for field, value in zip(model.__ommi_metadata__.fields.values(), row):
            if validator := self._find_type_validator(field.get("field_type", value)):
                yield validator(value)
            else:
                yield value

    def _find_type_validator(self, type_hint: Type[T]) -> Callable[[Any], T] | None:
        hint = get_origin(type_hint) or type_hint
        for validator_type, validator in self.type_validators.items():
            if issubclass(hint, validator_type):
                return validator

        return None

    def _get_sqlite_type(self, type_: Type) -> str:
        return self.type_mapping.get(type_, "TEXT")

    def _build_select_query(self, query: SelectQuery):
        query_builder = [f"SELECT * FROM {query.model.__ommi_metadata__.model_name}"]
        if query.where:
            query_builder.append(f"WHERE {query.where}")

        if query.limit > 0:
            query_builder.append(f"LIMIT {query.limit}")

            if query.offset > 0:
                query_builder.append(f"OFFSET {query.offset}")

        if query.order_by:
            ordering = ", ".join(
                f"{column} {'ASC' if ordering is ResultOrdering.ASCENDING else 'DESC'}"
                for column, ordering in query.order_by.items()
            )
            query_builder.append("ORDER BY")
            query_builder.append(ordering)

        return " ".join(query_builder) + ";"

    def _build_count_query(self, query: SelectQuery):
        query_builder = [
            f"SELECT Count(*) FROM {query.model.__ommi_metadata__.model_name}"
        ]
        if query.where:
            query_builder.append(f"WHERE {query.where}")

        if query.limit > 0:
            query_builder.append(f"LIMIT {query.limit}")

            if query.offset > 0:
                query_builder.append(f"OFFSET {query.offset}")

        return " ".join(query_builder) + ";"

    def _sync_with_last_inserted(self, item: OmmiModel, session: sqlite3.Cursor):
        pk = self._find_primary_key(type(item))
        result = session.execute("SELECT last_insert_rowid();").fetchone()
        setattr(item, pk, result[0])
